<?php

namespace Drupal\responsive_image_debugger\Entity;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Image\Image;
use Drupal\image\Entity\ImageStyle as ImageStyleBase;
use Drupal\image\ImageEffectInterface;
use Drupal\image\ImageEffectManager;

/**
 * Provides a custom image style class.
 *
 * We use this class to apply the text overlay on top of an image.
 */
class ImageStyle extends ImageStyleBase{

  const BG_COLOR = '#FF00FF';

  /**
   * Generate a semi-transparent image with custom color and text.
   *
   * @param string $color - Hex color (e.g., '#FF00FF')
   * @param string $text - Text to display on watermark
   * @param int $width - Width of the watermark image
   * @param int $height - Height of the watermark image
   *
   * @return string - Path to generated watermark image
   */
  public function generateWatermark(string $color, string $text, int $width, int $height, Image $drupal_image): string {

    $font_path = $this->getFont();
    $font_size = $this->getFontSize($drupal_image);
    $bbox = imagettfbbox($font_size, 0, $font_path, $text);
    $text_width = $bbox[2] - $bbox[0];
    $text_height = $bbox[1] - $bbox[7];
    $image = imagecreatetruecolor($text_width + 50, $text_height + 50);

    $text_color = imagecolorallocate($image, 0, 0, 0);
    imagesavealpha($image, true);
    $transparent_color = imagecolorallocatealpha($image, 0, 0, 0, 127);
    imagefill($image, 0, 0, $transparent_color);

    // Convert hex to RGBA
    [$r, $g, $b] = sscanf($color, "#%02x%02x%02x");
    $bg_color = imagecolorallocatealpha($image, $r, $g, $b, 80);
    imagefilledrectangle($image, 0, 0, $width, $height, $bg_color);

    // Add text
    $x = 20;
    $y = 50;

    imagettftext($image, $font_size, 0, (int) $x, (int) $y, $text_color, $font_path, $text);

    // Save watermark
    $filename = 'public://watermark_' . time() . '.png';
    imagepng($image, \Drupal::service('file_system')->realpath($filename));
    imagedestroy($image);

    return $filename;
  }

  /**
   * Apply watermark to the center of the original image.
   *
   * @param string $watermarkPath - Path to the watermark image
   * @param string $originalImagePath - Path to the original image
   *
   * @throws \Exception
   */
  public function applyWatermark(string $watermarkPath, string $originalImagePath): void {
    // Get image format from original file extension
    $ext = strtolower(pathinfo($originalImagePath, PATHINFO_EXTENSION));

    // Load original image
    switch ($ext) {
      case 'jpeg':
      case 'jpg':
        $original = imagecreatefromjpeg($originalImagePath);
        break;
      case 'png':
        $original = imagecreatefrompng($originalImagePath);
        break;
      case 'gif':
        $original = imagecreatefromgif($originalImagePath);
        break;
      default:
        throw new \Exception("Unsupported image format: $ext");
    }

    // Load watermark
    $watermark = imagecreatefrompng($watermarkPath);

    // Calculate position to center the watermark
    $orig_w = imagesx($original);
    $orig_h = imagesy($original);
    $wm_w = imagesx($watermark);
    $wm_h = imagesy($watermark);

    $x = ($orig_w - $wm_w) / 2;
    $y = ($orig_h - $wm_h) / 2;

    // Preserve alpha blending
    imagealphablending($original, true);
    imagesavealpha($original, true);

    // Merge watermark
    imagecopy($original, $watermark, (int) $x, (int) $y, 0, 0, $wm_w, $wm_h);

    // Save image in original format
    switch ($ext) {
      case 'jpeg':
      case 'jpg':
        imagejpeg($original, $originalImagePath, 90);
        break;
      case 'png':
        imagepng($original, $originalImagePath);
        break;
      case 'gif':
        imagegif($original, $originalImagePath);
        break;
    }

    // Clean up
    imagedestroy($original);
    imagedestroy($watermark);
  }

  /**
   * {@inheritdoc}
   */
  public function createDerivative($original_uri, $derivative_uri) {
    // If the source file doesn't exist, return FALSE without creating folders.
    $image = $this->getImageFactory()->get($original_uri);
    if (!$image->isValid()) {
      return FALSE;
    }

    // Get the folder for the final location of this style.
    $directory = \Drupal::service('file_system')->dirname($derivative_uri);

    // Build the destination folder tree if it doesn't already exist.
    if (!\Drupal::service('file_system')->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      \Drupal::logger('image')->error('Failed to create style directory: %directory', ['%directory' => $directory]);
      return FALSE;
    }

    foreach ($this->getEffects() as $effect) {
      $effect->applyEffect($image);
    }

    /** @var \Drupal\image_effects\Plugin\ImageEffect\TextOverlayImageEffect $image_effect */
    $image_effect = $this->buildDebugEffect($image);
    $effect_applied = $image_effect->applyEffect($image);

    if (!$image->save($derivative_uri)) {
      if (file_exists($derivative_uri)) {
        \Drupal::logger('image')->error('Cached image file %destination already exists. There may be an issue with your rewrite configuration.', ['%destination' => $derivative_uri]);
      }

      return FALSE;
    }

    if (!$effect_applied) {
      $watermark_path = $this->generateWatermark(self::BG_COLOR, $this->getText($image), 500, 200, $image);
      $this->applyWatermark($watermark_path, $derivative_uri);
    }

    return TRUE;
  }

  /**
   * Add imageapi_optimize module capability, in case it is enabled.
   *
   * @return string
   */
  public function getPipeline() {
    return '__default__';
  }

  /**
   * Builds and returns a debug effect configuration for an image.
   *
   * @param Image $image - The image object for which the debug effect is to be applied.
   *
   * @return ImageEffectInterface - The configured image effect instance.
   */
  private function buildDebugEffect(Image $image): ImageEffectInterface {

    $configuration = [
      'id' => 'image_effects_text_overlay',
      'weight' => -1000,
      'data' => [
        'text_string' => $this->getText($image),
        'font' => [
          'angle' => 0,
          'color' => '#000000FF',
          'name' => 'Montserrat Medium',
          'outline_bottom' => 0,
          'outline_left' => 0,
          'outline_right' => 0,
          'outline_top' => 0,
          'shadow_height' => 0,
          'shadow_width' => 0,
          'shadow_x_offset' => 1,
          'shadow_y_offset' => 1,
          'size' => $this->getFontSize($image),
          'stroke_color' => '#000000FF',
          'stroke_mode' => 'outline',
          'uri' => $this->getFont(),
        ],
        'text' => [
          'maximum_chars' => NULL,
          'align' => 'left',
          'case_format' => '',
          'decode_entities' => TRUE,
          'excess_chars_text' => '…',
          'fixed_width' => FALSE,
          'line_spacing' => 0,
          'maximum_width' => $image->getWidth() - 5,
          'strip_tags' => TRUE,
        ],
        'layout' => [
          'background_color' => self::BG_COLOR,
          'extended_color' => '',
          'overflow_action' => 'extend',
          'padding_bottom' => 10,
          'padding_left' => 10,
          'padding_right' => 10,
          'padding_top' => 10,
          'x_offset' => 0,
          'x_pos' => 'center',
          'y_offset' => 0,
          'y_pos' => 'center',
        ],
      ],
    ];

    /** @var ImageEffectManager $image_effect_manager */
    $image_effect_manager = \Drupal::service('plugin.manager.image.effect');
    return $image_effect_manager->createInstance('image_effects_text_overlay', $configuration);
  }

  /**
   * Retrieve the file path to the font used for rendering text.
   *
   * @return string - The full file path to the font file.
   */
  private function getFont(): string {
    $module_path = \Drupal::service('extension.list.module')
      ->getPath('responsive_image_debugger');
    return $module_path . '/fonts/MontserratMedium.otf';
  }

  /**
   * Determines the font size based on the width of the given image.
   *
   * @param Image $image - The image object used to calculate the font size.
   *
   * @return int - The calculated font size as an integer.
   */
  private function getFontSize(Image $image): int {
    $font_size = 24;
    if ($image->getWidth() >= 400) {
      $font_size = $image->getWidth() * .02;
    }
    elseif ($image->getWidth() <= 400 && $image->getWidth() >= 200) {
      $font_size = $image->getWidth() * .06;
    }
    elseif ($image->getWidth() <= 200) {
      $font_size = $image->getWidth() * .04;
    }
    return (int) $font_size;
  }

  /**
   * Generate a text description of the image properties.
   */
  private function getText(Image $image): string {
    $text_array = [
      'width: ' . $image->getWidth(),
      'height: ' . $image->getHeight(),
      "style name: \n" . $this->getName(),
    ];

    return implode("\n", $text_array);
  }


}
